import nltk
import pandas as pd
from nltk.corpus import webtext

nltk.download('webtext', quiet=True)
nltk.download('punkt', quiet=True)


def webtext_to_df(full_text: str) -> pd.DataFrame:
    # Separate the chat logs into individual messages
    text_rows = full_text.split("\n")
    # Create a dataframe with text rows
    df = pd.DataFrame({'text': text_rows})
    return df


def load_data_pirates():
    poc = webtext.raw('pirates.txt')
    return webtext_to_df(poc)


def load_data_king_arthur():
    king_arthur = webtext.raw('grail.txt')
    return webtext_to_df(king_arthur)


def load_data_chat_logs():
    chat_logs = webtext.raw('singles.txt')
    return webtext_to_df(chat_logs)



import pandas as pd

from tamtam.ab_info.ab_class import ABInfo


class BiasCalculation:
    def __init__(self, ab_info: ABInfo):
        self.ab_info = ab_info
        self.df = self.ab_info.df
        self.side_col = self.ab_info.test_info.get_side_col()[0]
        self.control = "A"
        self.treatment = "B"

    def calculate_bias(self):
        side_col = self.side_col
        id_cols = self.ab_info.test_info.get_id_cols()
        weight_col = self.ab_info.test_info.get_weight_col()
        df = self.df.copy()

        # Calculate count bias
        control_df = df.loc[df[side_col] == self.control]
        treatment_df = df.loc[df[side_col] == self.treatment]

        control_count = control_df.shape[0]
        treatment_count = treatment_df.shape[0]
        count_bias = treatment_count / control_count

        # Calculate ID bias
        control_ids = control_df[id_cols].drop_duplicates().shape[0]
        treatment_ids = treatment_df[id_cols].drop_duplicates().shape[0]
        id_bias = treatment_ids / control_ids

        # Calculate weight bias, if weight column exists
        if weight_col:
            control_weight = control_df[weight_col].sum()
            treatment_weight = treatment_df[weight_col].sum()
            weight_bias = treatment_weight / control_weight
        else:
            control_weight = control_count
            treatment_weight = treatment_count
            weight_bias = id_bias

        # Create a pandas dataframe with the bias values
        bias_df = pd.DataFrame({
            'bias_type': ['count', 'ids', 'weights'],
            'A': [control_count, control_ids, control_weight],
            'B': [treatment_count, treatment_ids, treatment_weight],
            'B/A - 1': [count_bias - 1, id_bias - 1, weight_bias - 1 if weight_bias else None]
        })

        return bias_df

    def run(self):
        print("Bias calculations")
        print(self.calculate_bias())



import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from chester.feature_stats.utils import round_value
from chester.zero_break.problem_specification import DataInfo


class NumericStats:
    def __init__(self, data_info: DataInfo, max_print=None):
        self.data_info = data_info
        self.max_print = max_print
        self.cols = self.data_info.feature_types_val["numeric"]
        self.data = self.data_info.data[self.cols]
        self.cols_sorted = self.sort_by_variance()

    def any_numeric(self):
        return True if len(self.cols) > 0 else False

    def sort_by_variance(self):
        if not self.any_numeric():
            return []
        variances = []
        for col in self.cols:
            data = self.data[col]
            variances.append((col, data.var()))
        sorted_variances = sorted(variances, key=lambda x: x[1], reverse=True)
        return [x[0] for x in sorted_variances]

    def plot_correlation(self, n=25, plot=True):
        if not self.any_numeric():
            return None
        if not plot:
            return None
        top_n = self.cols_sorted[:min(len(self.cols_sorted), 3 * n)]
        top_n_sampled = top_n[:min(len(top_n), n)]
        data = self.data.sort_values(by=top_n_sampled)
        corr = data[top_n_sampled].corr()
        if len(self.cols_sorted) <= n:
            plot_title = f"Pearson Correlation Plot"
        else:
            plot_title = f"Pearson Correlation Plot for {n} randomly sampled Features"
        plt.figure(figsize=(13, 13))
        plt.rcParams.update({'font.size': 18})
        sns.heatmap(corr, annot=False)
        plt.title(plot_title)
        print("Matrix correlation for numerical features")
        print("""\n
        💡 Rule of thumb:
        👍 Strong positive correlation: >= 0.7
        🤔 Moderate positive correlation: between 0.5 and 0.7
        🤨 Weak positive correlation: between 0.3 and 0.5
        🤷‍️ No/Negligible correlation: < 0.3
        """)
        plt.show()
        plt.close()

    def calculate_stats(self, is_print=True):
        from chester.util import ReportCollector, REPORT_PATH
        rc = ReportCollector(REPORT_PATH)

        if not self.any_numeric():
            return None
        result_dicts = []
        for col in self.cols:
            data = self.data[col]
            data_drop_dups = data.drop_duplicates()
            unique_values = data.nunique()
            missing_values = data.isnull().sum()
            max_vals = data_drop_dups.max()
            min_vals = data_drop_dups.min()
            avg_vals = data.mean()
            std_vals = data.std()
            n = len(data)
            ci_vals = [avg_vals - 1.645 * (std_vals / np.sqrt(n)), avg_vals + 1.645 * (std_vals / np.sqrt(n))]
            median_vals = data.median()
            top_vals = data_drop_dups.nlargest(3).apply(round_value)
            bottom_vals = data_drop_dups.nsmallest(3).apply(round_value)
            max_vals = max_vals.round(2)
            min_vals = min_vals.round(2)
            avg_vals = avg_vals.round(2)
            std_vals = std_vals.round(2)
            ci_vals = round_value(ci_vals[0]), round_value(ci_vals[1])
            median_vals = median_vals.round(2)
            top_vals = ",".join(map(str, top_vals.tolist()))
            bottom_vals = ",".join(map(str, bottom_vals.tolist()))
            result_dicts.append({'col': col, '# unique': unique_values, '# missing': missing_values,
                                 'max': max_vals,
                                 'min': min_vals, 'avg': avg_vals, 'std': std_vals, 'CI': ci_vals,
                                 'median': median_vals,
                                 'top_vals': top_vals, 'bottom_vals': bottom_vals})
        results_df = pd.DataFrame(result_dicts)
        if is_print:
            if self.max_print is not None:
                formatted_df = format_df(df=results_df,
                                         max_value_width=self.max_print,
                                         ci_max_value_width=self.max_print,
                                         col_max_value_width=self.max_print)
                print(formatted_df)
            else:
                formatted_df = format_df(results_df)
                print(formatted_df)
            len_df = len(formatted_df)
            rc.save_object(obj=formatted_df.sample(min(len_df, 10)), text="Feature stats:")
        return results_df

    def run(self, plot=True):
        self.calculate_stats()
        self.plot_correlation(plot=plot)
        return None


def format_df(df, max_value_width=25,
              col_max_value_width=25,
              ci_max_value_width=25,
              ci_col="CI", col_col="col"):
    pd.options.display.max_columns = None

    def trim_value(val):
        if len(str(val)) > max_value_width:
            return str(val)[:max_value_width] + "..."
        return str(val)

    def trim_ci_value(val):
        if len(str(val)) > ci_max_value_width:
            return str(val)[:ci_max_value_width] + "..."
        return str(val)

    def trim_col_value(val):
        if len(str(val)) > col_max_value_width:
            return str(val)[:ci_max_value_width] + "..."
        return str(val)

    df_subset = df.drop([ci_col, col_col], axis=1)
    df_subset = df_subset.applymap(trim_value)
    df[df_subset.columns] = df_subset
    df[ci_col] = df[ci_col].apply(trim_ci_value)
    df[col_col] = df[col_col].apply(trim_col_value)

    return df



from chester.model_training.data_preparation import CVData
from chester.zero_break.problem_specification import DataInfo


class BaseModel:
    def __init__(self, data_info: DataInfo, cv_data: CVData, num_models_to_compare=10, best_practice_prop=0.33):
        self.data_info = data_info
        self.cv_data = cv_data
        self.num_models_to_compare = num_models_to_compare
        self.best_practice_prop = best_practice_prop
        self.best_model = None
        self.best_metrics = {}
        try:
            self.feature_names = self.cv_data.train_data.columns
        except:
            pass

    def get_metric_function(self, metric):
        from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
        from sklearn.metrics import mean_squared_error, mean_absolute_error, mean_absolute_percentage_error
        from sklearn.metrics import roc_auc_score

        if metric == 'Accuracy':
            return accuracy_score
        elif metric == 'Precision':
            return precision_score
        elif metric == 'Recall':
            return recall_score
        elif metric == 'F1':
            return f1_score
        elif metric == "MSE":
            return mean_squared_error
        elif metric == "MAE":
            return mean_absolute_error
        elif metric == "MAPE":
            return mean_absolute_percentage_error
        elif metric == 'ROC':
            return roc_auc_score
        else:
            return None

    def get_metrics_functions(self):
        metric_functions = []
        metrics = self.data_info.metrics_detector_val
        for metric in metrics:
            metric_function = self.get_metric_function(metric)
            if metric_function is not None:
                metric_functions.append(metric_function)
        return metric_functions

    def conf_generator(self):
        pass

    def train_single_model(self, params: list):
        pass

    def get_best_model(self):
        pass



import numpy as np


def calculate_score_model(y, prediction):
    """
    Calculates evaluation metrics for the predictions
    :param y: The true labels
    :param prediction: predictions
    :return: A dictionary of metric scores for each model
    """
    from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
    import warnings
    warnings.filterwarnings('ignore')
    unique_values = len(np.unique(y))
    scores = {}
    if unique_values == 2:
        scores['accuracy_score'] = accuracy_score(y, prediction)
        scores['precision_score'] = precision_score(y, prediction)
        scores['precision_score'] = precision_score(y, prediction, zero_division=0)
        scores['recall_score'] = recall_score(y, prediction)
        scores['f1_score'] = f1_score(y, prediction)
    else:
        scores['accuracy_score'] = accuracy_score(y, prediction)
        scores['precision_score'] = precision_score(y, prediction, average='micro', zero_division=0)
        scores['recall_score'] = recall_score(y, prediction, average='micro')
        scores['f1_score'] = f1_score(y, prediction, average='micro')
    return scores



import pandas as pd
from sklearn.exceptions import UndefinedMetricWarning


def calculate_metric_score(y_true, y_pred, metric, problem_type_val):
    metric_name = metric.__name__
    if problem_type_val in ["Binary regression"]:
        y_pred = pd.Series(y_pred[:, 1], name='y_pred')
    elif problem_type_val in ["Multiclass classification", "Binary classification"]:
        y_pred = pd.Series(y_pred.argmax(axis=1), name='y_pred')
    try:
        return metric_name, metric(y_true, y_pred)
    except:
        try:
            if problem_type_val in ["Binary regression", "Binary classification"]:
                return metric_name, metric(y_true, y_pred, average='binary')
            elif problem_type_val == "Multiclass classification":
                return metric_name, metric(y_true, y_pred, average='macro')
        except:
            pass
    return None, None


def calculate_metrics_scores(y, prediction, metrics_list, problem_type=None):
    import warnings
    warnings.filterwarnings("ignore", category=UndefinedMetricWarning)
    results = {}
    for metric in metrics_list:
        metric_name, metric_value = calculate_metric_score(y, prediction, metric, problem_type)
        results[metric_name] = metric_value
    return results


from fuzzywuzzy import process


def is_metric_higher_is_better(metric_name: str) -> bool:
    higher_better_metrics = ["accuracy", "f1", "precision", "recall", "roc", "roc auc", "gini", "r squared", "mape",
                             "mae", "mse", "accuracy_score", "f1_score", "precision_score", "recall_score", "roc_score",
                             "roc_auc_score"]
    lower_better_metrics = ["rmse", "log loss", "cross entropy", "brier score", "loss", "mean_squared_error",
                            "mean_absolute_error", "mean_absolute_percentage_error"]

    match = process.extractOne(metric_name, higher_better_metrics + lower_better_metrics)
    if match[1] > 90:
        if match[0] in higher_better_metrics:
            return True
        elif match[0] in lower_better_metrics:
            return False
    else:
        raise ValueError(f"Metric {metric_name} not recognized.")



import random
import numpy as np
from diamond.data_augmentation.augmentation import ImageAugmentation
from diamond.image_data_info.image_info import ImageInfo
from diamond.model_training.best_model import ImageModelsTraining
from diamond.post_model_analysis.post_model import ImagePostModelAnalysis
from diamond.user_classes import ImagesData, ImagesAugmentationInfo, ImageModels, ImagePostModelSpec
import pandas as pd

from diamond.utils import index_labels


def run(images,
        image_shape,
        labels=None,
        validation_prop=0.3,
        is_augment_data=True, image_augmentation_info=None,
        is_train_model=True, image_models=None,
        is_post_model_analysis=True, image_post_model: ImagePostModelSpec = None,
        plot=True
        ):
    # Tell a story
    story = """Welcome to MadCat, the comprehensive machine learning and data analysis solution!
    \nThis module is designed to streamline the entire process image classification models, 
    \nfrom start to finish.
    \nTo learn more about MadCat, visit https://github.com/amito-ds/chester.\n"""
    print(story)
    diamond_collector = {}

    if labels is None:
        is_train_model = False
        labels = pd.Series([1] * len(images))
    # Image data
    if type(images).__name__ == "DataFrame":
        images = images.values

    label_dict, labels = index_labels(labels)

    image_data = ImagesData(images=images, labels=labels,
                            validation_prop=validation_prop,
                            image_shape=image_shape,
                            label_dict=label_dict)
    diamond_collector["image_data"] = image_data
    # plot
    if plot:
        print("Sample Plot")
        image_data.plot_images()

    # augmentation
    if is_augment_data:
        print("====> Augmenting data")
        if image_augmentation_info is None:
            image_augmentation_info = ImagesAugmentationInfo()
        image_data = ImageAugmentation(image_data, image_augmentation_info).run()
        diamond_collector["augmented_image_data"] = image_data
        # plot
        if plot:
            print("Updated Sample Plot")
            image_data.plot_images()

    # Training
    if not is_train_model:
        return diamond_collector
    if image_models is None:
        image_models = ImageModels()
    s = ''
    if image_models.n_models > 1:
        s = 's'
    print(f"====> Training {image_models.n_models} Model{s}")
    models_sorted = ImageModelsTraining(images_data=image_data,
                                        image_models=image_models).run()  # return the model ordered by best to worst

    diamond_collector["models"] = models_sorted

    # Post model analysis
    if is_post_model_analysis:
        print("====> Analyzing the best model")
        if image_post_model is None:
            image_post_model = ImagePostModelSpec(plot=plot)
        ImagePostModelAnalysis(model_list=models_sorted,
                               images_data=image_data,
                               image_post_model=image_post_model,
                               diamond_collector=diamond_collector,
                               plot=plot).run()

    return diamond_collector



def round_value(x):
    try:
        if abs(x) < 10:
            return round(x, 2)
        elif abs(x) < 100:
            return round(x, 1)
        else:
            return round(x)
    except:
        return None


from prettytable import PrettyTable


def create_pretty_table(df):
    # Convert the DataFrame to a list of dictionaries
    data = df.reset_index(drop=True).to_dict('records')

    # Create a PrettyTable object with column names
    table = PrettyTable(df.columns.tolist())

    # Add rows to the table
    for row in data:
        table.add_row([row[column] for column in df.columns])

    # Return the table
    return table



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import seaborn as sns
from sklearn import metrics

from chester.model_training.data_preparation import CVData


class ModelAnalyzer:
    def __init__(self, model):
        self.model = model

    def shap_values(self, X_train: pd.DataFrame):
        import shap
        explainer = shap.Explainer(self.model, X_train, check_additivity=False)
        shap_values = explainer(X_train, check_additivity=False)
        plt.title("SHAP values for train set")
        print(AnalyzeMessages().shap_values_message())
        shap.summary_plot(shap_values, X_train)

    def coefficients(self, X_train: pd.DataFrame) -> None:
        coef = self.model.coef_[0]
        feature_names = X_train.columns
        if len(coef) < 10:
            plt.bar(np.arange(len(coef)), coef)
            plt.xticks(np.arange(len(coef)), feature_names, rotation=90)
            plt.title("Coefficients for Logistic Regression Model")
            plt.xlabel("Features")
            plt.ylabel("Values")
            plt.show()
            plt.close()
        else:
            sns.violinplot(coef, inner="stick")
            plt.title("Coefficients distribution for logistic regression model")
            plt.xlabel("Coefficients")
            print(AnalyzeMessages().coefficients_message())
        plt.show()
        plt.close()

    def analyze(self, X_train: pd.DataFrame, y_train: pd.Series,
                X_test: pd.DataFrame, y_test: pd.Series,
                model, plot_shap_values: bool = True, coefficients: bool = True,
                confusion_matrix: bool = True,
                roc_curve: bool = True, learning_curve: bool = True,
                feature_importance: bool = True) -> None:
        unique_classes = len(np.unique(y_train))
        if feature_importance:
            try:
                self.plot_feature_importance(X_train)
            except:
                try:
                    self.plot_simple_feature_importance(X_train)
                except:
                    pass
        if plot_shap_values:
            if unique_classes == 2:
                try:
                    self.shap_values(X_train)
                except:
                    pass
        if coefficients:
            try:
                self.coefficients(X_train)
            except:
                pass
        if confusion_matrix:
            try:
                self.confusion_matrix(X_test, y_test)
            except:
                pass
        if roc_curve:
            if unique_classes == 2:
                self.roc_curve(X_test, y_test)
        if learning_curve:
            self.learning_curve(X_train, y_train)

    def confusion_matrix(self, X_test: pd.DataFrame, y_test: pd.Series) -> None:
        from sklearn.metrics import confusion_matrix
        import seaborn as sns
        print(AnalyzeMessages().confusion_matrix_message())

        y_pred = self.model.predict(X_test)
        cm = confusion_matrix(y_test, y_pred)
        sns.heatmap(cm, annot=True, cmap='Blues', fmt='g')
        plt.title('Confusion Matrix')
        plt.xlabel('Predicted')
        plt.ylabel('Actual')
        plt.show()
        plt.close()

    def roc_curve(self, X_test: pd.DataFrame, y_test: pd.Series) -> None:
        from sklearn.metrics import roc_curve, auc
        try:
            y_pred = self.model.predict_proba(X_test)
            fpr, tpr, thresholds = roc_curve(y_test, y_pred[:, 1])
            roc_auc = auc(fpr, tpr)
            plt.figure()
            plt.plot(fpr, tpr, color='darkorange', lw=1, label='ROC curve (area = %0.2f)' % roc_auc)
            plt.plot([0, 1], [0, 1], color='navy', lw=1, linestyle='--')
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.05])
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title('Receiver operating characteristic')
            plt.legend(loc="lower right")
            print(AnalyzeMessages().roc_curve_message())
            plt.show()
            plt.close()
        except:
            return None

    def learning_curve(self, X: pd.DataFrame, y: pd.Series) -> None:
        try:
            from sklearn.model_selection import learning_curve
            train_sizes, train_scores, test_scores = learning_curve(self.model, X, y, cv=5, scoring='accuracy')
            train_scores_mean = np.mean(train_scores, axis=1)
            train_scores_std = np.std(train_scores, axis=1)
            test_scores_mean = np.mean(test_scores, axis=1)
            test_scores_std = np.std(test_scores, axis=1)
            plt.plot(train_sizes, train_scores_mean, 'o-', color="r", label="Training score")
            plt.plot(train_sizes, test_scores_mean, 'o-', color="g", label="Cross-validation score")
            plt.fill_between(train_sizes, train_scores_mean - train_scores_std,
                             train_scores_mean + train_scores_std, alpha=0.1, color="r")
            plt.fill_between(train_sizes, test_scores_mean - test_scores_std,
                             test_scores_mean + test_scores_std, alpha=0.1, color="g")
            plt.grid()
            plt.xlabel("Training examples")
            plt.ylabel("Accuracy Score")
            plt.title("Learning Curve")
            plt.legend(loc="best")
            print(AnalyzeMessages().learning_curve_message())
            plt.show()
            plt.close()
        except:
            plt.close()
            return None

    def plot_simple_feature_importance(self, X_train: pd.DataFrame):
        feature_importance = self.model.feature_importances_
        feature_importance = 100.0 * (feature_importance / feature_importance.max())
        feature_names = X_train.columns
        important_idx = np.argsort(feature_importance)
        data = {'feature_names': feature_names[important_idx], 'feature_importance': feature_importance[important_idx]}
        df = pd.DataFrame(data)
        sns.barplot(y='feature_names', x='feature_importance', data=df)
        plt.xlabel("Feature importance")
        plt.show()
        plt.close()

    def plot_feature_importance(self, X_train: pd.DataFrame):
        feature_importance = self.model.feature_importances_
        feature_importance = 100.0 * (feature_importance / feature_importance.max())
        feature_names = X_train.columns
        important_idx = np.argsort(feature_importance)
        data = {'feature_names': feature_names[important_idx], 'feature_importance': feature_importance[important_idx]}
        df = pd.DataFrame(data)
        print(AnalyzeMessages().feature_importance_message())
        fig = px.bar(df, x='feature_importance', y='feature_names', orientation='h', text='feature_importance')
        fig.show()
        plt.close()


class AnalyzeMessages:
    def shap_values_message(self):
        return "SHAP values can be used to understand the importance of each feature in the model's" \
               " predictions.\n The plot shows the average absolute SHAP value of each feature for all " \
               "the samples in the test set.\n Features with higher absolute SHAP values have a greater " \
               "impact on the model's predictions.\n"

    def coefficients_message(self):
        return "The coefficient plot shows the weight of each feature in the model's predictions.\n" \
               " Positive coefficients indicate that the feature has a positive relationship with" \
               " the target variable\n, while negative coefficients indicate a negative relationship.\n" \
               " The size of the boxplot represents the range of values for each feature.\n"

    def performance_metrics_message(self):
        return "The performance metrics table shows the results of different evaluation metrics" \
               " applied on the model's predictions\n. These metrics can give you an idea of how well" \
               " the model is performing in terms of accuracy, precision, recall, and other measures.\n"

    def confusion_matrix_message(self):
        return "The confusion matrix shows the number of true positive, " \
               "true negative, false positive, and false negative predictions made by the model.\n " \
               "This can give you an idea of how well the model is able to distinguish between the " \
               "different classes.\n"

    def roc_curve_message(self):
        return "The ROC curve shows the trade-off between true positive rate (sensitivity) " \
               "and false positive rate (1-specificity) for different threshold settings.\n T" \
               "he AUC (Area under the curve) value gives an overall measure of the model's performance.\n"

    def learning_curve_message(self):
        return "The learning curve displays model performance as training samples increase.\n" \
               "High training and low validation suggest overfitting,\n" \
               "low training and high validation suggest underfitting.\n"

    def feature_importance_message(self):
        return "The feature importance plot shows the relative importance of each feature in the model's predictions.\n" \
               " Features with higher importance have a greater impact on the model's predictions and are more useful\n" \
               " for making accurate predictions.\n"


def analyze_model(model, cv_data: CVData, target_label='target'):
    X_train, X_test, y_train, y_test = \
        cv_data.train_data.drop(columns=[target_label]), \
            cv_data.test_data.drop(columns=[target_label]), \
            cv_data.train_data[target_label], \
            cv_data.test_data[target_label]
    analyzer = ModelAnalyzer(model)
    analyzer.analyze(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test,
                     model=model)


def get_default_metrics(y):
    n_classes = len(np.unique(y))
    if n_classes == 2:
        return [metrics.accuracy_score, metrics.f1_score, metrics.mean_squared_error, metrics.r2_score]
    elif n_classes > 2:
        return [metrics.accuracy_score, metrics.f1_score]
    else:
        return [metrics.mean_squared_error, metrics.r2_score]


import Levenshtein


def get_traffic_light(metric_name, value):
    thresholds = {"Accuracy": (0.8, 0.9, 0.95),
                  "F1-Score": (0.7, 0.8, 0.9),
                  "Precision": (0.7, 0.8, 0.9),
                  "Recall": (0.7, 0.8, 0.9),
                  "ROC-AUC": (0.7, 0.8, 0.9),
                  "BLEU Score": (0.7, 0.8, 0.9),
                  "METEOR Score": (0.7, 0.8, 0.9),
                  "Rouge Score": (0.7, 0.8, 0.9),
                  "CIDEr Score": (0.7, 0.8, 0.9),
                  "Embedding Average Cosine Similarity": (0.7, 0.8, 0.9)}
    metric_name = metric_name.lower()
    closest_metric = None
    closest_distance = float('inf')
    for metric in thresholds.keys():
        distance = Levenshtein.distance(metric_name, metric.lower())
        if distance < closest_distance:
            closest_metric = metric
            closest_distance = distance
    if closest_metric is None:
        return 'grey'
    if value < thresholds[closest_metric][0]:
        return 'red'
    elif value < thresholds[closest_metric][1]:
        return 'yellow'
    elif value < thresholds[closest_metric][2]:
        return 'green'
    else:
        return 'bright green'



import numpy as np

from chester.zero_break.problem_specification import DataInfo


def min_fix(l):
    if not l:
        return None
    return min(l)


def max_fix(l):
    if not l:
        return None
    return max(l)


def mean_fix(l):
    if not l:
        return None
    return np.average(l)


def median_fix(l):
    if not l:
        return None
    return np.median(l)



from typing import Tuple

import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer


def get_tfidf_embedding(train_df: pd.DataFrame,
                        text_column: str = 'text',
                        test_df: pd.DataFrame = None,
                        ngram_range: Tuple[int, int] = (1, 1), embedding_size=100):
    """
    Creates a TF-IDF embedding for the text data in the given DataFrames.
    Parameters:
    - train_df: A DataFrame with the training data.
    - test_df: A DataFrame with the test data.
    - text_column: The name of the column in the DataFrames that contains the text data.
    - ngram_range: The range of n-grams to consider when creating the embedding.

    Returns:
    - A tuple with the following elements:
        - A DataFrame with the TF-IDF embedding for the training data.
        - A DataFrame with the TF-IDF embedding for the test data (if provided).
        - The TfidfVectorizer object used to create the embedding.
    """

    # Create the TfidfVectorizer object
    vectorizer = TfidfVectorizer(ngram_range=ngram_range, max_features=embedding_size)

    # Fit the vectorizer on the training data
    X_train = vectorizer.fit_transform(train_df[text_column])

    # Get the feature names
    feature_names = list(vectorizer.vocabulary_.keys())

    # Create the embedding DataFrame for the training data
    embedding_train_df = pd.DataFrame(X_train.toarray(), columns=[f"tfidf_{word}" for word in feature_names])

    # Check if test data was provided
    if test_df is not None:
        # Transform the test data using the vectorizer
        X_test = vectorizer.transform(test_df[text_column])

        # Create the embedding DataFrame for the test data
        embedding_test_df = pd.DataFrame(X_test.toarray(), columns=[f"tfidf_{word}" for word in feature_names])

        return embedding_train_df, embedding_test_df, vectorizer
    else:
        return embedding_train_df, None, vectorizer



import re
import string
import unicodedata
from typing import List

import pandas as pd

from chester.util import get_stopwords, ReportCollector, REPORT_PATH


def remove_punctuation(text: str) -> str:
    """Remove all punctuation from the given text."""
    return str(text).translate(str(text).maketrans("", "", string.punctuation))


def remove_numbers(text: str) -> str:
    """Remove all numbers from the given text."""
    return ''.join(c for c in text if not c.isdigit())


def remove_whitespace(text: str) -> str:
    """Remove excess whitespace from the given text."""
    return ' '.join(str(text).split())


def remove_empty_lines(text: str):
    """Remove excess empty lines from the given text."""
    lines = str(text).splitlines()
    cleaned_text = '\n'.join([line for line in lines if line.strip()])
    return cleaned_text


def lowercase(text: str) -> str:
    """Convert the given text to lowercase."""
    return text.lower()


def remove_stopwords(text: str, stopwords: List[str]) -> str:
    """Remove common words that do not contribute to the meaning of the text.

    stopwords: a list of words to remove from the text.
    """
    words = text.split()
    cleaned_words = [word for word in words if word not in stopwords]
    return ' '.join(cleaned_words)


def remove_accented_characters(text: str) -> str:
    """Remove accented characters from the given text."""
    return ''.join(c for c in unicodedata.normalize('NFKD', text) if not unicodedata.combining(c))


def remove_special_characters(text: str) -> str:
    """Remove special characters from the given text."""
    return re.sub(r'[^\w\s]', '', text)


def remove_html_tags(text: str) -> str:
    """Remove HTML tags from the given text."""
    return re.sub(r'<[^<]+?>', '', text)


def clean_df_text(text_column, cleaning_options: dict):
    print_cleaning_options(**cleaning_options)
    return text_column.apply(lambda x: clean_text(x, **cleaning_options))


def print_cleaning_options(
        stopwords: str,
        remove_punctuation_flag: bool = True,
        remove_numbers_flag: bool = True,
        remove_whitespace_flag: bool = True,
        lowercase_flag: bool = True,
        remove_stopwords_flag: bool = False,
        remove_accented_characters_flag: bool = True,
        remove_special_characters_flag: bool = True,
        remove_html_tags_flag: bool = True):
    if remove_punctuation_flag:
        print("Removing punctuation")
    if remove_numbers_flag:
        print("Removing numbers")
    if remove_whitespace_flag:
        print("Removing whitespaces")
    if lowercase_flag:
        print("Lowercasing text")
    if remove_stopwords_flag:
        print("Removing stop words")
    if remove_accented_characters_flag:
        print("Removing accented characters")
    if remove_special_characters_flag:
        print("Removing special characters")
    if remove_html_tags_flag:
        print("Removing html tags")


def clean_text(text: str,
               remove_punctuation_flag: bool = True,
               remove_numbers_flag: bool = True,
               remove_whitespace_flag: bool = True,
               remove_empty_line_flag: bool = True,
               lowercase_flag: bool = True,
               remove_stopwords_flag: bool = True,
               stopwords: List[str] = None,
               remove_accented_characters_flag: bool = True,
               remove_special_characters_flag: bool = True,
               remove_html_tags_flag: bool = True) -> str:
    """Apply a series of cleaning functions to the given text.

    text: the text to clean.
    remove_punctuation_flag: a flag indicating whether to remove punctuation from the text.
    remove_numbers_flag: a flag indicating whether to remove numbers from the text.
    remove_whitespace_flag: a flag indicating whether to remove excess whitespace from the text.
    remove_empty_line_flag: a flag indicating whether to remove excess empty lines from the text.
    lowercase_flag: a flag indicating whether to convert the text to lowercase.
    remove_stopwords_flag: a flag indicating whether to remove common words that do not contribute to the meaning of the text.
    stopwords: a list of words to remove from the text. Required if remove_stopwords_flag is True.
    remove_accented_characters_flag: a flag indicating whether to remove accented characters from the text.
    remove_special_characters_flag: a flag indicating whether to remove special characters from the text.
    remove_html_tags_flag: a flag indicating whether to remove HTML tags from the text.
    """

    text = str(text)
    if remove_punctuation_flag:
        text = remove_punctuation(text)
    if remove_numbers_flag:
        text = remove_numbers(text)
    if remove_whitespace_flag:
        text = remove_whitespace(text)
    if remove_empty_line_flag:
        remove_empty_lines(text)
    if lowercase_flag:
        text = lowercase(text)
    if remove_stopwords_flag:
        if not stopwords:
            stopwords = get_stopwords()
        text = remove_stopwords(text, stopwords)
    if remove_accented_characters_flag:
        text = remove_accented_characters(text)
    if remove_special_characters_flag:
        text = remove_special_characters(text)
    if remove_html_tags_flag:
        text = remove_html_tags(text)
    return text


from typing import List


class TextCleaner:
    def __init__(self,
                 df: pd.DataFrame = None,
                 text_column: str = 'text',
                 remove_punctuation_flag: bool = True,
                 remove_numbers_flag: bool = True,
                 remove_whitespace_flag: bool = True,
                 remove_empty_line_flag: bool = True,
                 lowercase_flag: bool = True,
                 remove_stopwords_flag: bool = True,
                 stopwords: List[str] = None,
                 remove_accented_characters_flag: bool = True,
                 remove_special_characters_flag: bool = True,
                 remove_html_tags_flag: bool = True):
        self.df = df
        self.text_column = text_column
        self.remove_punctuation_flag = remove_punctuation_flag
        self.remove_numbers_flag = remove_numbers_flag
        self.remove_whitespace_flag = remove_whitespace_flag
        self.remove_empty_line_flag = remove_empty_line_flag
        self.lowercase_flag = lowercase_flag
        self.remove_stopwords_flag = remove_stopwords_flag
        self.stopwords = stopwords
        self.remove_accented_characters_flag = remove_accented_characters_flag
        self.remove_special_characters_flag = remove_special_characters_flag
        self.remove_html_tags_flag = remove_html_tags_flag

    def generate_report(self):
        rc = ReportCollector(REPORT_PATH)
        report_str = ""
        if self.remove_punctuation_flag:
            report_str += "Removing punctuation, "
        if self.remove_numbers_flag:
            report_str += "Removing numbers, "
        if self.remove_whitespace_flag:
            report_str += "Removing whitespaces, "
        if self.remove_empty_line_flag:
            report_str += "Removing empty lines, "
        if self.lowercase_flag:
            report_str += "Lowercasing text, "
        if self.remove_stopwords_flag:
            if self.stopwords:
                report_str += f"Stopwords are given, "
            else:
                report_str += "Removing stopwords, "
        if self.remove_accented_characters_flag:
            report_str += "Removing accented characters, "
        if self.remove_special_characters_flag:
            report_str += "Removing special characters, "
        if self.remove_html_tags_flag:
            report_str += "Removing html tags, "
        if report_str:
            report_str = report_str[:-2]
            title_to_print = f"The following cleaning steps will be applied to clean column '{self.text_column}': {report_str}."
            print(title_to_print)
            rc.save_text(title_to_print)
        else:
            print("No cleaning steps selected.")


def clean_text_df(text_cleaner: TextCleaner) -> pd.DataFrame:
    df = text_cleaner.df
    text_col = text_cleaner.text_column
    df[text_col] = df[text_col].apply(lambda x: clean_text(
        x,
        remove_punctuation_flag=text_cleaner.remove_punctuation_flag,
        remove_numbers_flag=text_cleaner.remove_numbers_flag,
        remove_whitespace_flag=text_cleaner.remove_whitespace_flag,
        remove_empty_line_flag=text_cleaner.remove_empty_line_flag,
        lowercase_flag=text_cleaner.lowercase_flag,
        remove_stopwords_flag=text_cleaner.remove_stopwords_flag,
        stopwords=text_cleaner.stopwords,
        remove_accented_characters_flag=text_cleaner.remove_accented_characters_flag,
        remove_special_characters_flag=text_cleaner.remove_special_characters_flag,
        remove_html_tags_flag=text_cleaner.remove_html_tags_flag
    ))
    return df



import math
import pickle
import pandas as pd
import numpy as np
import os
from sklearn.preprocessing import LabelEncoder
# exec(open("/Users/amitosi/PycharmProjects/chester/diamond/manual_run.py").read())
from PIL import Image

from diamond.run import run
from diamond.user_classes import ImageModels, ImageModel, ImagesAugmentationInfo
from sklearn.datasets import fetch_openml

with open("/Users/amitosi/PycharmProjects/chester/chester/data/data_batch_1", 'rb') as fo:
    data = pickle.load(fo, encoding='bytes')
# Extract the image and label data from the batch


# Load the MNIST dataset
mnist = fetch_openml('mnist_784')
# Extract the data and labels
X, y = mnist['data'], mnist['target']

# X = X.astype(np.float32)
# X.to_csv('mnist_fashion_X.csv', index=False)
# y.to_csv('mnist_fashion_y.csv', index=False)


image_model_list = [
    ImageModel(network_name="EfficientNetB0",
               batch_size=64*64*16,
               num_epochs=1,
               optimizer_params={'lr': 0.005},
               dropout=0.7),

    # ImageModel(network_name="ResNet101", batch_size=64 * 2, num_epochs=8, optimizer_params={'lr': 0.05},
    #            dropout=0.5),
    #
    # ImageModel(network_name="ResNet101", batch_size=64 * 2, num_epochs=6, optimizer_params={'lr': 0.05},
    #            dropout=0.2),
    # ImageModel(network_name="EfficientNetB0", batch_size=32 * 32, num_epochs=5, optimizer_params={'lr': 1})
]
image_models = ImageModels(image_model_list=image_model_list)

# y = pd.read_csv('/Users/amitosi/PycharmProjects/chester/mnist_fashion_y.csv')
# y = y['class']
# X = pd.read_csv('/Users/amitosi/PycharmProjects/chester/mnist_fashion_X.csv')

image_shape = (28, 28)
# image_shape = (32, 32)
diamond_collector = run(images=X,
                        image_shape=image_shape,
                        labels=y,
                        is_augment_data=True,
                        image_augmentation_info=ImagesAugmentationInfo(aug_prop=0.2),
                        is_train_model=True, image_models=image_models,
                        is_post_model_analysis=True,
                        plot=True)

# labels handling
# image convert to np nd array
#
# images = data[b'data'][0:2500]
# # labels = np.array(data[b'labels'])[0:2500]
# labels = data[b'labels'][0:2500]
# image_shape = (3, 32, 32)
# #
# diamond_collector = run(images=images,
#                         image_shape=image_shape,
#                         labels=labels,
#                         is_augment_data=True,
#                         image_augmentation_info=ImagesAugmentationInfo(aug_prop=0.01),
#                         is_train_model=True, image_models=image_models,
#                         is_post_model_analysis=True,
#                         plot=True)

# print(diamond_collector["models"])
#
#

#
# # define directory path
# data_dir = '/Users/amitosi/PycharmProjects/chester/chester/data/weather'
#
# # # define image size
# img_size = (64, 64)
# # img_size = (32, 32)
#
# # define empty arrays to store images and labels
# images = []
# labels = []
#
# # loop over the directories containing the images
#
# # loop over the files in the directory
# for filename in os.listdir(data_dir):
#     if filename.endswith('.jpg'):
#         label = filename.split("/")[-1].split(".")[0]
#         label = ''.join(char for char in label if not char.isdigit())
#         # read the image as a PIL.Image object
#         img = Image.open(os.path.join(data_dir, filename)).convert('RGB')
#         # print(img.size)
#         # resize the image
#         img = img.resize(img_size)
#         # convert the image to a NumPy array
#         img_array = np.array(img)
#         # append the image array to the list of images
#         images.append(img_array)
#         # append the label to the list of labels
#         labels.append(label)
#
# # # convert the lists to NumPy arrays
# images = np.array(images)
# # print(np.unique(labels))
# labels = np.array(labels)
# # # create a LabelEncoder object
# # encoder = LabelEncoder()
# # #
# # # fit the encoder to the labels and transform the labels to integer indices
# # labels = encoder.fit_transform(labels)
# # #
# image_shape = (3, 64, 64)
# # # image_shape = (3, 640, 480)
#
# diamond_collector = run(images=images,
#                         image_shape=image_shape,
#                         labels=labels,
#                         is_augment_data=True,
#                         image_augmentation_info=ImagesAugmentationInfo(aug_prop=0.1),
#                         is_train_model=True, image_models=image_models,
#                         is_post_model_analysis=True,
#                         plot=True)



import pandas as pd
from sklearn.model_selection import train_test_split

from chester.features_engineering.bag_of_words import get_bow_embedding
from chester.features_engineering.corex import get_corex_embedding
from chester.features_engineering.tfidf import get_tfidf_embedding
from chester.util import ReportCollector, REPORT_PATH


def get_embeddings(training_data: pd.DataFrame,
                   test_data: pd.DataFrame = None,
                   split_data: bool = True, split_prop: float = 0.3,
                   split_random_state=42,
                   text_column="clean_text", target_column='target',
                   corex=True, corex_dim=50, tfidf=True,
                   tfidf_dim=100, bow=True, bow_dim=100,
                   ngram_range=(1, 1)):
    rc = ReportCollector(REPORT_PATH)
    rc.save_text("Extracting embedding")
    if split_data:
        if not test_data:
            training_data, test_data = train_test_split(training_data, test_size=split_prop,
                                                        random_state=split_random_state)

    order_list = ["First", "Next", "Then", "Additionally", "Furthermore", "Then", "Additionally", "Furthermore", "Then",
                  "Additionally"]
    # Define empty DataFrames for the embeddings
    corex_embedding = pd.DataFrame()
    tfidf_embedding = pd.DataFrame()
    bow_embedding = pd.DataFrame()
    ner_bow_embedding = pd.DataFrame()
    corex_test_embedding = pd.DataFrame()
    tfidf_test_embedding = pd.DataFrame()
    bow_test_embedding = pd.DataFrame()
    ner_bow_test_embedding = pd.DataFrame()
    i = 0
    # Extract Corex topic model embeddings if requested
    if corex:
        title_to_print = f"{order_list[i]}, Extracting Corex topic model embeddings with dimension {corex_dim}"
        print(title_to_print)
        rc.save_text(title_to_print)

        corex_embedding, corex_test_embedding = get_corex_embedding(training_data=training_data, test_data=test_data,
                                                                    ngram_range=ngram_range,
                                                                    n_topics=corex_dim, text_column=text_column)
        i += 1

    # Extract TF-IDF embeddings if requested
    if tfidf:
        title_to_print = f"{order_list[i]}, Extracting TF-IDF embeddings with dimension {tfidf_dim}"
        print(title_to_print)
        rc.save_text(title_to_print)
        tfidf_embedding, tfidf_test_embedding, _ = get_tfidf_embedding(training_data, test_df=test_data,
                                                                       ngram_range=ngram_range,
                                                                       embedding_size=tfidf_dim,
                                                                       text_column=text_column)
        i += 1

    # Extract bag-of-words embeddings if requested
    if bow:
        title_to_print = f"{order_list[i]}, Extracting bag-of-words embeddings with dimension {bow_dim}"
        print(title_to_print)
        rc.save_text(title_to_print)
        bow_embedding, bow_test_embedding, _ = get_bow_embedding(training_data, test_data=test_data,
                                                                 ngram_range=ngram_range,
                                                                 embedding_size=bow_dim,
                                                                 text_column=text_column)
        i += 1

    # Concatenate the embeddings and return them
    embeddings = pd.concat([corex_embedding, tfidf_embedding, ner_bow_embedding, bow_embedding], axis=1)
    test_embeddings = pd.concat(
        [corex_test_embedding, tfidf_test_embedding, ner_bow_test_embedding, bow_test_embedding], axis=1)

    title_to_print = f"Lastly, All embeddings have been concatenated"
    print(title_to_print)
    rc.save_text(title_to_print)
    # adding the label to train and test embedding
    try:
        training_data.reset_index(drop=True, inplace=True)
        embeddings.reset_index(drop=True, inplace=True)
        embeddings = pd.concat([embeddings, training_data[target_column]], axis=1)

        test_data.reset_index(drop=True, inplace=True)
        test_embeddings.reset_index(drop=True, inplace=True)
        test_embeddings = pd.concat([test_embeddings, test_data[target_column]], axis=1)
    except:
        pass
    return embeddings, test_embeddings


class TextFeatureExtraction:
    def __init__(self, training_data: pd.DataFrame = None,
                 test_data: pd.DataFrame = None,
                 split_data: bool = True, split_prop: float = 0.3, split_random_state=42,
                 text_column="text", target_column='target',
                 corex=True, corex_dim=50, tfidf=True, tfidf_dim=100, bow=True, bow_dim=100,
                 ngram_range=(1, 1)):
        self.training_data = training_data
        self.test_data = test_data
        self.split_data = split_data
        self.split_prop = split_prop
        self.split_random_state = split_random_state
        self.text_column = text_column
        self.target_column = target_column
        self.corex = corex
        self.corex_dim = corex_dim
        self.tfidf = tfidf
        self.tfidf_dim = tfidf_dim
        self.bow = bow
        self.bow_dim = bow_dim
        self.ngram_range = ngram_range


def extract_features(feature_extractor: TextFeatureExtraction):
    training_data = feature_extractor.training_data
    test_data = feature_extractor.test_data
    split_data = feature_extractor.split_data
    split_prop = feature_extractor.split_prop
    split_random_state = feature_extractor.split_random_state
    text_column = feature_extractor.text_column
    target_col = feature_extractor.target_column
    corex = feature_extractor.corex
    corex_dim = feature_extractor.corex_dim
    tfidf = feature_extractor.tfidf
    tfidf_dim = feature_extractor.tfidf_dim
    bow = feature_extractor.bow
    bow_dim = feature_extractor.bow_dim
    ngram_range = feature_extractor.ngram_range

    return get_embeddings(training_data, test_data, split_data, split_prop, split_random_state, text_column, target_col,
                          corex, corex_dim, tfidf, tfidf_dim, bow, bow_dim, ngram_range)



from tamtam.ab_info.ab_class import ABInfo
from tamtam.user_class.user_class import TestInfo


class ABPartialPlot:
    def __init__(self, ab_info: ABInfo, test_info: TestInfo, metric):
        self.ab_info = ab_info
        self.test_info = test_info
        self.metric = metric
        self.df = self.ab_info.df

    def run(self):
        pass



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import torch
from IPython import get_ipython
from sklearn.metrics import confusion_matrix, accuracy_score
from sklearn.metrics import precision_score, recall_score

from diamond.model_training.model_results import ImageModelResults
from diamond.user_classes import ImagesData, ImagePostModelSpec


class ImagePostModelAnalysis:
    def __init__(self, model_list,
                 images_data: ImagesData,
                 image_post_model: ImagePostModelSpec,
                 plot=True,
                 diamond_collector=None):
        self.model_list = model_list
        self.images_data = images_data
        self.image_post_model = image_post_model
        self.plot = plot
        self.best_model = self.model_list[0].model
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.label_dict = self.images_data.label_dict

        self.diamond_collector = {} if diamond_collector is None else diamond_collector
        self.network_parameters = self.model_list[0].network_parameters
        self.image_model_training = self.model_list[0].image_model_training
        self.train_loader, self.val_loader = self.images_data.create_data_loaders(
            batch_size=self.network_parameters["_batch_size"])
        self.eval_predictions = self.get_prediction()
        self.eval_labels = self.images_data.labels_val

    def get_prediction(self):
        predicted = self.image_model_training.evaluate_predictions
        self.diamond_collector["eval predictions"] = predicted
        return predicted

    def compare_models(self):
        data = []
        for model in self.model_list:
            data.append({**model.network_parameters,
                         "train_loss": model.train_loss, "val_loss": model.val_loss})
        df = pd.DataFrame(data)
        code = """diamond_collector["models comparison"]"""
        print(f"Comparing for all Models is Saved, for exploration, run {code}")
        self.diamond_collector["models comparison"] = df
        return df

    @staticmethod
    def replace_predictions(eval_labels_unique, eval_predictions):
        updated_predictions = []
        for eval_prediction in eval_predictions:
            if eval_prediction not in eval_labels_unique:
                updated_predictions.append(-1)
            else:
                updated_predictions.append(eval_prediction)
        return updated_predictions

    def map_labels(self, x):
        for k, v in self.label_dict.items():
            if v == x:
                return k
        return x  # Return the original value if it is not in the label_dict dictionary

    # def confusion_matrix(self, plot):
    #     if "class" not in self.images_data.problem_type:
    #         return None
    #
    #     vmap_labels = np.vectorize(self.map_labels)
    #     eval_labels = vmap_labels(self.eval_labels)
    #     eval_predictions = vmap_labels(self.eval_predictions)
    #     try:
    #         cm = confusion_matrix(eval_labels, eval_predictions)
    #     except:
    #         cm = confusion_matrix(eval_labels, eval_labels.cpu().numpy())
    #     self.diamond_collector["confusion matrix"] = cm
    #     # Create confusion matrix plot
    #     if plot:
    #         fig, ax = plt.subplots(figsize=(8, 8))
    #         sns.heatmap(cm, annot=True, fmt='d', cmap=plt.cm.Blues, ax=ax, cbar=False)
    #         ax.set_xlabel('Predicted label')
    #         ax.set_ylabel('True label')
    #         ax.set_title('Confusion Matrix')
    #         plt.show()
    #     return cm

    def confusion_matrix(self, plot):
        if "class" not in self.images_data.problem_type:
            return None

        vmap_labels = np.vectorize(self.map_labels)
        eval_labels = vmap_labels(self.eval_labels)
        eval_predictions = vmap_labels(self.eval_predictions)
        try:
            cm = confusion_matrix(eval_labels, eval_predictions)
        except:
            cm = confusion_matrix(eval_labels, eval_labels.cpu().numpy())

        # Get the actual class names in sorted order
        sorted_class_names = sorted(self.label_dict, key=lambda x: self.label_dict[x])

        # Create a heatmap plot with class names as tick labels
        if plot:
            fig, ax = plt.subplots(figsize=(8, 8))
            sns.heatmap(cm, annot=True, fmt='d', cmap=plt.cm.Blues, ax=ax, cbar=False,
                        xticklabels=sorted_class_names, yticklabels=sorted_class_names)
            ax.set_xlabel('Predicted label')
            ax.set_ylabel('True label')
            ax.set_title('Confusion Matrix')
            plt.show()

        self.diamond_collector["confusion matrix"] = cm
        return cm

    def precision_recall(self):
        import warnings
        warnings.filterwarnings("ignore", message="invalid value encountered in divide")

        if "class" not in self.images_data.problem_type:
            return None
        vmap_labels = np.vectorize(self.map_labels)
        eval_labels = vmap_labels(self.eval_labels)
        eval_predictions = vmap_labels(self.eval_predictions)
        # Calculate precision and recall for each class
        try:
            precision = precision_score(eval_labels, eval_predictions, average=None)
            recall = recall_score(eval_labels, eval_predictions, average=None)
            accuracy = accuracy_score(eval_labels, eval_predictions)
        except:
            eval_pred = eval_predictions.cpu().numpy()
            precision = precision_score(eval_labels, eval_pred, average=None)
            recall = recall_score(eval_labels, eval_pred, average=None)
            accuracy = accuracy_score(eval_labels, eval_pred)

        f1_score = 2 * (precision * recall) / (precision + recall)

        # Create a dataframe with the results
        label_values = np.unique(eval_labels)
        results = pd.DataFrame({"Class": label_values,
                                "Accuracy": accuracy,
                                "Precision": precision,
                                "Recall": recall,
                                "F1-Score": f1_score})

        # Display the results
        self.diamond_collector["Precision Recall"] = results
        return results

    def run(self):
        if self.image_post_model.is_compare_models:
            self.compare_models()
        if self.image_post_model.is_confusion_matrix:
            self.confusion_matrix(plot=self.plot)
        if self.image_post_model.is_precision_recall:
            self.precision_recall()



from tamtam.ab_feature_analysis.ab_feature_analysis_class import ABFeatureAnalysis
from tamtam.ab_info.ab_class import ABInfo
from tamtam.allocation_calculation.bias_class import BiasCalculation
from tamtam.delta_analysis.delta_class import DeltaPValueCalc
from tamtam.metrics_correlation.metric_class import MetricCorrelation
from tamtam.user_class.user_class import ABData, TestInfo


def run(ab_data: ABData, test_info: TestInfo):
    # AB info
    ab_info = ABInfo(ab_data=ab_data, test_info=test_info)

    # calculate bias
    print("==========================> Bias calculations: ")
    BiasCalculation(ab_info).run()

    # metrics corr
    MetricCorrelation(ab_info).run()

    # Delta, Pvalue
    print("==========================>  Test Analysis ")
    DeltaPValueCalc(ab_info=ab_info, test_info=test_info).run()

    # Features
    if len(ab_info.test_info.get_feature_cols()) > 0:
        ABFeatureAnalysis(ab_info=ab_info, test_info=test_info).run()



from typing import List

import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn.metrics import accuracy_score, precision_recall_fscore_support, recall_score, f1_score
from sklearn.preprocessing import LabelBinarizer
from sklearn.preprocessing import LabelEncoder

from chester.model_training.data_preparation import CVData, Parameter


# from tensorflow.keras.layers import LSTM, Dense


def get_default_parameters(X_train: pd.DataFrame):
    default_parameters = {
        'input_shape': (X_train.shape[1], 1),
        'output_dim': 10,
        'recurrent_dropout': 0.1,
        'dropout': 0.1,
        'optimizer': 'adam',
        'loss': 'binary_crossentropy',
        'metrics': ['accuracy']
    }
    lstm_default_parameters = [
        Parameter(name, default_parameters[name]) for name in default_parameters
    ]

    return lstm_default_parameters


def train_lstm(X_train, y_train, parameters):
    """
    Trains a LSTM model using the given parameters.
    :param X_train: The training data features
    :param y_train: The training data labels
    :param parameters: A list of dictionaries, each representing a set of hyperparameters
    :return: A list of trained LSTM models
    """
    if not parameters:
        parameters = get_default_parameters(X_train)
    params = {}
    for param in parameters:
        params[param.name] = param.value
    model = tf.keras.Sequential()
    model.add(
        tf.keras.layers.LSTM(params['output_dim'], input_shape=params['input_shape'],
                             recurrent_dropout=params['recurrent_dropout'],
                             dropout=params['dropout']))
    model.add(tf.keras.layers.Dense(1, activation='sigmoid'))
    model.compile(loss='binary_crossentropy', optimizer=params['optimizer'], metrics=['accuracy'])

    # convert y_train to numerical values
    label_encoder = LabelEncoder()
    y_train = label_encoder.fit_transform(y_train)
    model.fit(X_train, y_train)
    return model, label_encoder


def predict_lstm(model, X):
    """
    Makes predictions using a list of LSTM models
    :param model: trained LSTM model
    :param X: The data to make predictions on
    :return: A list of predictions for each model
    """
    return model.predict(X)


def score_lstm(y, prediction, metric_funcs: List[callable]):
    """
    Calculates evaluation metrics for the predictions
    :param y: The true labels
    :param prediction: predictions
    :param metric_funcs: A list of evaluation metric functions
    :return: A dictionary of metric scores for each model
    """
    label_binarizer = LabelBinarizer()
    y = label_binarizer.fit_transform(y)
    pred = label_binarizer.transform(prediction)
    scores = {metric.__name__: metric(y, pred) for metric in metric_funcs}
    return scores


def lstm_with_outputs(cv_data: CVData, parameters, target_col: str,
                      metric_funcs: List[callable] = None, label_encoder=None):
    results = []
    if not metric_funcs:
        metric_funcs = [accuracy_score, precision_recall_fscore_support, recall_score, f1_score]
    if not parameters:
        parameters = get_default_parameters(cv_data.train_data.drop(columns=[target_col]))
    for p in parameters:
        print(p.name, p.value)
    pass
    for i, (train_index, test_index) in enumerate(cv_data.splits):
        X_train, X_test = cv_data.train_data.iloc[train_index], cv_data.train_data.iloc[test_index]
        y_train, y_test = X_train[target_col], X_test[target_col]
        X_train = X_train.drop(columns=[target_col])
        X_test = X_test.drop(columns=[target_col])
        # reshape input data to 3D
        X_train = X_train.to_numpy()
        X_train = X_train.reshape(X_train.shape[0], X_train.shape[1], 1)
        X_test = X_test.to_numpy()
        X_test = X_test.reshape(X_test.shape[0], X_test.shape[1], 1)
        model, label_encoder = train_lstm(X_train, y_train, parameters)
        train_pred = predict_lstm(model, X_train)
        test_pred = predict_lstm(model, X_test)
        train_pred = np.round(train_pred)
        test_pred = np.round(test_pred)
        train_scores = score_lstm(y_train, train_pred, metric_funcs)
        test_scores = score_lstm(y_test, test_pred, metric_funcs)
        results.append(
            {'type': 'train', 'fold': i, **{param.name: param.value for param in parameters}, **train_scores})
        results.append({'type': 'test', 'fold': i, **{param.name: param.value for param in parameters}, **test_scores})
    model, label_encoder = train_lstm(cv_data.train_data.drop(columns=[target_col]), cv_data.train_data[target_col],
                                      parameters)
    return results, model, parameters, model.predict(cv_data.train_data.drop(columns=[target_col]))



def calculate_metric_score(y_true, y_pred, metric, problem_type_val):
    metric_name = metric.__name__
    try:
        # print("trying to", metric(y_true, y_pred))
        return metric_name, metric(y_true, y_pred)
    except:
        try:
            if problem_type_val in ["Binary regression", "Binary classification"]:
                return metric_name, metric(y_true, y_pred, average='binary')
            elif problem_type_val == "Multiclass classification":
                return metric_name, metric(y_true, y_pred, average='macro')
        except:
            pass
    return None, None


def calculate_metrics_scores(y, prediction, metrics_list, problem_type=None):
    results = {}
    for metric in metrics_list:
        metric_name, metric_value = calculate_metric_score(y, prediction, metric, problem_type)
        results[metric_name] = metric_value
    return results



from collections import Counter

import matplotlib.pyplot as plt
import seaborn as sns

from chester.util import ReportCollector, REPORT_PATH


def most_common_words(data, common_words=10, text_column='text'):
    rc = ReportCollector(REPORT_PATH)
    # create a Counter object to store the word counts
    word_counts = Counter()

    # iterate over each row in the data
    for index, row in data.iterrows():
        # split the text into a list of words
        words = row[text_column].split()

        # add the words to the Counter
        word_counts.update(words)

    # sort the Counter by value (count) in descending order
    sorted_word_counts = sorted(word_counts.items(), key=lambda x: x[1], reverse=True)

    # create a barplot using seaborn
    plt.figure(figsize=(11, 11))
    plt.rcParams.update({'font.size': 25})
    rc.save_text(str(sorted_word_counts[:100]))  # save top 100 common words
    sns.barplot(x=[t[0] for t in sorted_word_counts[:common_words]],
                y=[t[1] for t in sorted_word_counts[:common_words]])
    plt.xlabel('Word')
    plt.ylabel('Count')
    plt.title('Most Common Words')
    plt.show()
    plt.close()

    # return the sorted list of tuples (word, count)
    return sorted_word_counts[:common_words]



from collections import Counter

import numpy as np
import pandas as pd
from nltk.tokenize import sent_tokenize
from sklearn.decomposition import TruncatedSVD
from sklearn.feature_extraction.text import TfidfVectorizer

from chester.util import ReportCollector, REPORT_PATH


def extract_key_sentences_lsa(text, k=10):
    """
    Extract k most important sentences from text using LSA algorithm.
    :param text: the text to extract key sentences from
    :param k: the number of sentences to extract (default: 10)
    :return: list of k most important sentences
    """
    # tokenize the text into sentences
    sentences = sent_tokenize(text)

    # apply tf-idf weighting to the sentences
    vectorizer = TfidfVectorizer()
    X = vectorizer.fit_transform(sentences)

    # apply LSA to the tf-idf weighted sentences
    svd = TruncatedSVD(n_components=k)
    X = svd.fit_transform(X)

    # use cosine similarity to identify the most important sentences
    similarity = np.asarray(np.asmatrix(X) * np.asmatrix(X).T)
    sentence_scores = similarity.diagonal()

    # sort the sentences by importance
    sentence_scores_sorted = sentence_scores.argsort()[::-1]
    return_list = [sentences[i] for i in sentence_scores_sorted]
    return return_list


def key_sentences(df, text_column='text', common_sentences=10):
    # Concatenate all text in the specified column into one string
    all_text = ". ".join(df[text_column])

    # Tokenize the string into sentences
    sentences = sent_tokenize(all_text)

    # Create a Counter to store the frequency of each sentence
    cnt = Counter()
    for sentence in sentences:
        cnt[sentence] += 1

    # return the most common sentences
    return cnt.most_common(common_sentences)


def extract_key_sentences(df: pd.DataFrame, text_column='text', algorithm='LSA', n_sentences=10):
    """
    Extract k most important sentences from a pandas dataframe containing text data.
    :param df: the dataframe to extract key sentences from
    :param text_column: the name of the column containing the text data (default: 'clean_text')
    :param algorithm: the algorithm to use for sentence extraction (default: 'summarization')
    :param n_sentences: the number of sentences to extract (default: 10)
    :return: list of k most important sentences
    """
    rc = ReportCollector(REPORT_PATH)

    full_text = '. '.join(df[text_column])
    if algorithm == 'LSA':
        sentences = list(set(extract_key_sentences_lsa(full_text, n_sentences)))[0:n_sentences]
    elif algorithm == 'common_sentences':
        tokenized_sentences = [sentence.split() for sentence in sent_tokenize(full_text)]
        sentence_counter = Counter([tuple(sentence) for sentence in tokenized_sentences])
        common_sentences = [" ".join(sentence) for sentence, count in sentence_counter.most_common(n_sentences)]
        sentences = common_sentences
    else:
        raise ValueError(f'Invalid algorithm: {algorithm}')

    title = f'\tKey sentences out of based on {algorithm} algorithm: {sentences} \n'
    rc.save_text(title)
    return title


def extract_first_k_words(text, k):
    """
    Extract the first k words from text.
    :param text: the text to extract words from
    :param k: the number of words to extract
    :return: list of k first words in the text
    """
    words = text.split()  # split text into words using the whitespace as a delimiter
    return ' '.join(words[:k])  # return the first k words



import random
from typing import List

import pandas as pd
from sklearn.exceptions import UndefinedMetricWarning

from chester.model_training.data_preparation import CVData
from chester.model_training.data_preparation import Parameter
from chester.model_training.models.chester_models.base_model_utils import is_metric_higher_is_better
from chester.model_training.models.chester_models.hp_generator import HPGenerator
from chester.model_training.models.chester_models.linear_regression.linear_regression import LinearRegressionModel
from chester.zero_break.problem_specification import DataInfo


def train_linear_regression(X_train, y_train, parameters: list, data_info: DataInfo):
    """
    Trains a baseline model using the given parameters.
    :param X_train: The training data features (unused in this function)
    :param y_train: The training data labels
    :param parameters: parameters list
    :return: A trained baseline model
    """
    model = LinearRegressionModel(parameters, data_info)
    model.fit(X_train, y_train)
    return model


def predict_linear_regression(model, X):
    """
    Makes predictions using a baseline model
    :param model: trained baseline model
    :param X: The data to make predictions on (unused in this function)
    :return: A list of predictions
    """
    return model.transform(X)


def linear_regression_with_outputs(cv_data: CVData,
                                   metrics: list,
                                   target_col: str,
                                   parameters: list,
                                   data_info: DataInfo,
                                   ):
    results = []
    for i, (X_train, y_train, X_test, y_test) in enumerate(cv_data.format_splits()):
        model = train_linear_regression(X_train, y_train, parameters=parameters, data_info=data_info)
        prediction = predict_linear_regression(model, X_test)
        prediction_train = predict_linear_regression(model, X_train)
        scores = calculate_linear_regression_metrics_scores(y_test, prediction, metrics, data_info.problem_type_val)
        results.append({'type': 'test', 'fold': i, **scores})
        scores = calculate_linear_regression_metrics_scores(y_train, prediction_train, metrics,
                                                            data_info.problem_type_val)
        results.append({'type': 'train', 'fold': i, **scores})
    return results, model


def compare_models(results):
    all_results = [(pd.DataFrame(result), model) for result, model in results]
    metric_name = [col for col in all_results[0][0].columns if col not in ['type', 'fold']][0]
    sort_ascending = is_metric_higher_is_better(metric_name)
    best_result = None
    best_model = None
    best_value = None
    for (result, model) in all_results:
        test_result = result[result['type'] == 'test'].groupby('fold').mean(numeric_only=True).reset_index()
        mean_value = test_result[metric_name].mean()
        if best_value is None or \
                (sort_ascending and mean_value > best_value) \
                or (not sort_ascending and mean_value < best_value):
            best_value = mean_value
            best_result = result
            best_model = model
    return best_result, best_model


default_parameters = {}


def generate_linear_regression_configs(k: int = 10, best_practice_prob=0.33) -> List[List[Parameter]]:
    hp_generator = HPGeneratoEN(n_models=k, best_practice_prob=best_practice_prob)
    parameter_format = hp_generator.hp_format(hp_generator.generate_configs())
    return parameter_format


class HPGeneratoEN(HPGenerator):
    def __init__(self, best_practice_configs: list = None,
                 categorical_sample_configs: list = None,
                 n_models=9,
                 best_practice_prob=0.33):
        super().__init__(best_practice_configs, categorical_sample_configs, n_models, best_practice_prob)
        self.best_practice_configs = self.load_best_practice_configs()

    @staticmethod
    def load_best_practice_configs():
        return [
            {**default_parameters, 'alpha': 0, 'l1_ratio': 0},  # linear
            {**default_parameters, 'alpha': 0.1, 'l1_ratio': 0},  # ridge
            {**default_parameters, 'alpha': 0.1, 'l1_ratio': 1},  # lasso
            {**default_parameters, 'alpha': 0.1, 'l1_ratio': 0.1},
            {**default_parameters, 'alpha': 0.1, 'l1_ratio': 0.5},
            {**default_parameters, 'alpha': 0.1, 'l1_ratio': 0.7},
            {**default_parameters, 'alpha': 0.1, 'l1_ratio': 0.9},
            {**default_parameters, 'alpha': 0.1, 'l1_ratio': 1.0},
            {**default_parameters, 'alpha': 0.5, 'l1_ratio': 0.1},
            {**default_parameters, 'alpha': 0.5, 'l1_ratio': 0.5},
            {**default_parameters, 'alpha': 0.5, 'l1_ratio': 0.7},
            {**default_parameters, 'alpha': 0.5, 'l1_ratio': 0.9},
            {**default_parameters, 'alpha': 1, 'l1_ratio': 0},  # ridge
            {**default_parameters, 'alpha': 1, 'l1_ratio': 1},  # lasso
            {**default_parameters, 'alpha': 0.5, 'l1_ratio': 1.0},
            {**default_parameters, 'alpha': 1.0, 'l1_ratio': 0.1},
            {**default_parameters, 'alpha': 1.0, 'l1_ratio': 0.5},
            {**default_parameters, 'alpha': 1.0, 'l1_ratio': 0.7},
            {**default_parameters, 'alpha': 1.0, 'l1_ratio': 0.9}
        ]

    def generate_random_config(self) -> dict:
        method = random.choices(['Elastic net', 'Lasso', 'Ridge'], weights=[0.5, 0.25, 0.25])[0]
        if method == 'Elastic net':
            alpha = random.uniform(0, 1)
            l1_ratio = random.uniform(0, 1)
        elif method == 'Lasso':
            alpha = random.uniform(0, 1)
            l1_ratio = 1
        elif method == 'Ridge':
            alpha = random.uniform(0, 1)
            l1_ratio = 0
        return {**default_parameters, 'alpha': alpha, 'l1_ratio': l1_ratio}


def calculate_linear_regression_metric_score(y_true, y_pred, metric, problem_type_val):
    import warnings
    warnings.filterwarnings("ignore", category=UndefinedMetricWarning)
    metric_name = metric.__name__
    try:
        return metric_name, metric(y_true, y_pred)
    except:
        try:
            if problem_type_val in ["Binary regression", "Binary classification"]:
                return metric_name, metric(y_true, y_pred, average='binary')
            elif problem_type_val == "Multiclass classification":
                return metric_name, metric(y_true, y_pred, average='macro')
        except:
            pass
    return None, None


def calculate_linear_regression_metrics_scores(y, prediction, metrics_list, problem_type=None):
    results = {}
    for metric in metrics_list:
        metric_name, metric_value = calculate_linear_regression_metric_score(y, prediction, metric, problem_type)
        results[metric_name] = metric_value
    return results



import pandas as pd

from chester.run.user_classes import TimeSeriesHandler
from chester.zero_break.problem_specification import DataInfo


class TSStaticFeatures:
    def __init__(self, column,
                 col_name,
                 time_series_handler: TimeSeriesHandler = None,
                 data_info: DataInfo = None):
        self.col_name = col_name
        self.column = column
        self.time_series_handler = time_series_handler
        self.data_info = data_info
        self.df = self.data_info.data
        self.df[self.col_name] = pd.to_datetime(self.df[self.col_name])  # conver to datetime

    def extract_minute(self):
        """
        Extract the minute in the day from a datetime column.
        :param df: pandas DataFrame
        :param datetime_col: string, the name of the datetime column
        :return: pandas series
        """
        return self.df[self.col_name].dt.hour * 60 + self.df[self.col_name].dt.minute, "ts_minutes_" + self.col_name

    def extract_hour(self):
        """
        Extract the hour from the datetime column.
        :return: pandas series
        """
        return self.df[self.col_name].dt.hour, "ts_hours_" + self.col_name

    def extract_day_in_month(self):
        """
        Extract the day of the month from the datetime column.
        :return: pandas series
        """
        return self.df[self.col_name].dt.day, "ts_dayinmonth_" + self.col_name

    def extract_day_in_week(self):
        """
        Extract the day of the week from the datetime column.
        :return: pandas series
        """
        return self.df[self.col_name].dt.strftime("%A"), "ts_dayinweek_" + self.col_name

    def extract_month_in_year(self):
        """
        Extract the month of the year from the datetime column.
        :return: pandas series
        """
        return self.df[self.col_name].dt.month, "ts_month_" + self.col_name

    def extract_quarter(self):
        """
        Extract the quarter of the year from the datetime column.
        :return: pandas series
        """
        return self.df[self.col_name].dt.quarter, "ts_quarter_" + self.col_name

    def extract_year(self):
        """
        Extract the year from the datetime column.
        :return: pandas series
        """
        return self.df[self.col_name].dt.year, "ts_year_" + self.col_name

    def run(self):
        # call all feature extraction methods and concatenate the resulting Pandas series
        minute, minute_name = self.extract_minute()
        hour, hour_name = self.extract_hour()
        day, day_name = self.extract_day_in_month()
        weekday, weekday_name = self.extract_day_in_week()
        month, month_name = self.extract_month_in_year()
        quarter, quarter_name = self.extract_quarter()
        year, year_name = self.extract_year()
        ts_feat_names = [minute_name, hour_name, day_name, weekday_name, month_name, quarter_name, year_name]
        ts_features = pd.concat([minute, hour, day, weekday, month, quarter, year], axis=1)
        ts_features.columns = ts_feat_names
        return ts_features, ts_feat_names
#
# create a sample DataFrame with a datetime column
# data = pd.DataFrame({
#     "date": pd.date_range(start="2022-01-22 07:00:000", periods=10, freq="D"),
#     "value": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
# })
#
# data = pd.DataFrame({
#     "date": pd.to_datetime(["2022-01-22 07:00:00", "2022-01-23 07:00:00", None, "2022-01-25 07:00:00", None, "2022-01-27 07:00:00", "2022-01-28 07:00:00", "2022-01-29 07:00:00", "2022-01-30 07:00:00", "2022-01-31 07:00:00"]),
#     "value": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
# })
#
# # create a DataInfo object with the sample DataFrame
# data_info = DataInfo(data=data, target="value")
#
# # create a TSStaticFeatures object and call the run method to extract time-related features
# static_features = TSStaticFeatures(
#     column=data.date,
#     col_name="date",
#     time_series_handler=None,
#     data_info=data_info
# )
# features, feature_names = static_features.run()
#
# # print the extracted features and feature names
# print(features[0:3].T)
# print(feature_names)
